#include "MainWindow.hpp"
#include "core.hpp"
#include <gtkmm/signallistitemfactory.h>
#include <Database/entities.hpp>
#include <vector>
#include <string>
#include <YoutubeApi/youtubeapi.hpp>
#include <gdkmm/texture.h>
#include <gtkmm/picture.h>
#include <gtkmm/label.h>
#include <gdkmm/pixbuf.h>
#include <glibmm.h>

#include <iostream>

namespace components {

    MainWindow* MainWindow::create(core::Core* appcore)
    {
        auto builder = Gtk::Builder::create_from_resource("/src/YTMPV/res/MainWindow.ui");
        auto win = Gtk::Builder::get_widget_derived<MainWindow>(
            builder,
            "mainWindow",
            appcore);
        assert(win != nullptr);
        return win;
    }
    
    MainWindow::MainWindow(
        BaseObjectType* cobject,
        const Glib::RefPtr<Gtk::Builder>& refBuilder,
        core::Core* appcore)
        // init
        : Gtk::ApplicationWindow(cobject)
        , m_appcore(appcore)
        , m_builder(refBuilder)
        , m_rate_container(Gtk::Orientation::HORIZONTAL)
    {
        // Загрузка виджетов
        auto video_list = m_builder->get_widget<Gtk::ListView>("videoList");
        auto search_button = m_builder->get_widget<Gtk::Button>("searchButton");
        auto start_search_button = m_builder->get_widget<Gtk::Button>("startSearchButton");
        m_search_body = m_builder->get_widget<Gtk::Paned>("searchBody");
        m_search_entry = m_builder->get_widget<Gtk::Entry>("searchEntry");
        m_start_search_entry = m_builder->get_widget<Gtk::Entry>("startSearchEntry");
        m_stackswitcher = m_builder->get_widget<Gtk::StackSwitcher>("switcher");
        m_body = m_builder->get_widget<Gtk::Stack>("body");

        assert(video_list != nullptr);
        assert(search_button != nullptr);
        assert(start_search_button != nullptr);
        assert(m_search_body != nullptr);
        assert(m_search_entry != nullptr);
        assert(m_start_search_entry != nullptr);
        assert(m_stackswitcher != nullptr);
        assert(m_body != nullptr);
        
        // -- Создание модели выбора видео в поиске --
        m_video_storage = Gio::ListStore<components::VideoModel>::create();
        m_selection_model = Gtk::SingleSelection::create(m_video_storage);
        m_selection_model->set_autoselect(false);
        m_selection_model->set_can_unselect(true);
        m_selection_model->property_selected().signal_changed().connect(
            sigc::mem_fun(*this, &MainWindow::onVideoChanged));
        video_list->set_model(m_selection_model);

        // -- Создание фабрики производства элементов списка видео --
        auto factory = Gtk::SignalListItemFactory::create();
        factory->signal_setup().connect(
            sigc::mem_fun(*this, &MainWindow::onVideoSetup));
        factory->signal_bind().connect(
            sigc::mem_fun(*this, &MainWindow::onVideoBind));
        video_list->set_factory(factory);

        // -- Подключение обработчиков к кнопкам --

        // Кнопки поиска
        search_button->signal_clicked().connect(
            sigc::mem_fun(*this, &MainWindow::onSearchButtonClicked));
        start_search_button->signal_clicked().connect(
            sigc::mem_fun(*this, &MainWindow::onStartSearchButtonClicked));

        // TODO брать из настроек
        m_search_body->set_position(256);

        // Кнопки оценки видео
        m_rate_dislike_btn = Gtk::make_managed<Gtk::Button>("Dislike");
        m_rate_meh_btn = Gtk::make_managed<Gtk::Button>("No rating");
        m_rate_like_btn = Gtk::make_managed<Gtk::Button>("Like");
        m_rate_container.set_homogeneous();
        m_rate_container.append(*m_rate_dislike_btn);
        m_rate_container.append(*m_rate_meh_btn);
        m_rate_container.append(*m_rate_like_btn);

        m_rate_dislike_btn->signal_clicked().connect(sigc::bind(
            sigc::mem_fun(*this, &MainWindow::onRateButtonClicked),
            RateButtonParams { core::VideoRating::DISLIKE }));
        m_rate_meh_btn->signal_clicked().connect(sigc::bind(
            sigc::mem_fun(*this, &MainWindow::onRateButtonClicked),
            RateButtonParams { core::VideoRating::NONE } ));
        m_rate_like_btn->signal_clicked().connect(sigc::bind(
            sigc::mem_fun(*this, &MainWindow::onRateButtonClicked),
            RateButtonParams { core::VideoRating::LIKE } ));
    }

    void MainWindow::searchVideoByQuery(Glib::ustring query)
    {
        // Очистить предыдущие результаты поиска
        m_video_storage->remove_all();

        // Получить новые результаты поиска
        db::video found;
        ytapi::SearchResults response = m_appcore->searchVideos(query);

        if (response.status == ytapi::Status::NET_ERROR) {
            searchSidebarError("A net error happened.\nCheck your internet connection.");
            return;
        }

        if (response.status == ytapi::Status::APIKEY_INVALID) {
            searchSidebarError("Provided invalid api key.");
            return;
        }

        // Добавить в список видео найденные видео
        for (auto ytid : response.ytids) {
            found = m_appcore->getVideoByYTID(ytid);
            m_video_storage->append(VideoModel::create(found));
        }
        searchSidebarRegularMode();
    }

    void MainWindow::onSearchButtonClicked()
    {
        auto query = m_search_entry->get_text();
        searchVideoByQuery(query);
    }

    void MainWindow::onStartSearchButtonClicked()
    {
        auto query = m_start_search_entry->get_text();
        m_search_entry->set_text(query);
        m_body->set_visible_child(
            "videos",
            Gtk::StackTransitionType::SLIDE_LEFT
        );
        searchVideoByQuery(query);
    }

    void MainWindow::onVideoSetup(const Glib::RefPtr<Gtk::ListItem>& list_item)
    {
        // Главный контейнер
        auto layout = Gtk::make_managed<Gtk::Box>(Gtk::Orientation::HORIZONTAL);

        // Превью
        auto thumbnail = Gtk::make_managed<Gtk::Picture>();

        // Контейнер деталей видео
        auto details = Gtk::make_managed<Gtk::Box>(Gtk::Orientation::VERTICAL);
        details->set_margin(8);

        // Название
        auto title = Gtk::make_managed<Gtk::Label>();
        title->set_halign(Gtk::Align::START);

        // Контейнер для времени, канала и кол-ва просмотров
        auto info = Gtk::make_managed<Gtk::Box>(Gtk::Orientation::HORIZONTAL);
        info->set_homogeneous();
        info->set_hexpand();

        // Надпись: время публикации
        auto published_at_label = Gtk::make_managed<Gtk::Label>();
        published_at_label->set_halign(Gtk::Align::START);

        // Надпись: название канала
        auto channel_label = Gtk::make_managed<Gtk::Label>();
        channel_label->set_halign(Gtk::Align::START);

        // Надпись: количество просмотров
        auto views_count_label = Gtk::make_managed<Gtk::Label>();
        views_count_label->set_halign(Gtk::Align::START);

        // Контейнер для кнопок
        auto buttons = Gtk::make_managed<Gtk::Box>(Gtk::Orientation::HORIZONTAL);
        buttons->set_spacing(4);
        buttons->set_margin(8);

        // Кнопка "Воспроизвести"
        auto play_button = Gtk::make_managed<Gtk::Button>();
        play_button->set_icon_name("media-playback-start");

        // Кнопка "Добавить в очередь"
        auto queue_button = Gtk::make_managed<Gtk::Button>();
        queue_button->set_icon_name("list-add");

        info->append(*published_at_label);
        info->append(*channel_label);
        info->append(*views_count_label);

        buttons->append(*play_button);
        buttons->append(*queue_button);

        details->append(*title);
        details->append(*info);
        details->append(*buttons);

        layout->append(*thumbnail);
        layout->append(*details);
        
        list_item->set_child(*layout);
    }

    void MainWindow::onVideoBind(const Glib::RefPtr<Gtk::ListItem>& list_item)
    {
        auto pos = list_item->get_position();
        if (pos == GTK_INVALID_LIST_POSITION) return;
        auto item = m_video_storage->get_item(pos);
        auto obj = item->m_obj;
        
        auto layout = dynamic_cast<Gtk::Box*>(list_item->get_child());
        if (!layout) return;

        auto thumbnail = dynamic_cast<Gtk::Picture*>(layout->get_first_child());
        if (!thumbnail) return;

        auto details = dynamic_cast<Gtk::Box*>(thumbnail->get_next_sibling());
        if (!details) return;

        auto title = dynamic_cast<Gtk::Label*>(details->get_first_child());
        if (!title) return;

        auto info = dynamic_cast<Gtk::Box*>(title->get_next_sibling());
        if (!info) return;

        auto published_at_label = dynamic_cast<Gtk::Label*>(info->get_first_child());
        if (!published_at_label) return;

        auto channel_label = dynamic_cast<Gtk::Label*>(published_at_label->get_next_sibling());
        if (!channel_label) return;

        auto views_count_label = dynamic_cast<Gtk::Label*>(channel_label->get_next_sibling());
        if (!views_count_label) return;

        auto buttons = dynamic_cast<Gtk::Box*>(info->get_next_sibling());
        if (!buttons) return;

        auto play_button = dynamic_cast<Gtk::Button*>(buttons->get_first_child());
        if (!play_button) return;

        auto queue_button = dynamic_cast<Gtk::Button*>(play_button->get_next_sibling());
        if (!queue_button) return;

        // Загрузка изображения превью видео с диска
        // TODO: ~/.cache/ytmpv/thumbnails -- брать из appcore
        std::string thumbnail_path = "thumbnails/default/" + obj.ytid + ".jpg";
        Glib::RefPtr<Gdk::Pixbuf> thumbnail_pixbuf;
        try {
            thumbnail_pixbuf = Gdk::Pixbuf::create_from_file(thumbnail_path, 120, 90);
        } catch (Glib::FileError& e) {
            switch (e.code()) {
                case Glib::FileError::Code::NO_SUCH_ENTITY:
                    // Превью на диске не найдено. Скачиваем...
                    ytapi::downloadMediumThumnail(obj.ytid, thumbnail_path);
                    // ...и присваиваем
                    thumbnail_pixbuf = Gdk::Pixbuf::create_from_file(thumbnail_path, 120, 90);
                    break;

                default:
                    std::cout << "Unhandled error when creating thumbnail. Code is: " << e.code(); 
                    break;
            }
        }
        auto thumbnail_texture = Gdk::Texture::create_for_pixbuf(thumbnail_pixbuf);
        thumbnail->set_paintable(thumbnail_texture);
        thumbnail->set_can_shrink(false);
        thumbnail->set_content_fit(Gtk::ContentFit::COVER);

        title->set_text(obj.title);

        published_at_label->set_text(m_appcore->getDateText(obj.published_at));
        channel_label->set_text(obj.author_obj.name);
        views_count_label->set_text(m_appcore->getNumberGrouped(obj.views_count));
    }

    void MainWindow::onVideoChanged()
    {
        auto position = m_selection_model->get_selected();
        auto item = m_video_storage->get_item(position);
        if (!item) return;
        auto obj = item->m_obj;

        searchSidebarVideoDetailsMode(
            obj.title,
            obj.author_obj.name,
            obj.description,
            obj.published_at,
            obj.ytid,
            obj.views_count,
            obj.big_thumbnail
        );
    }

    void MainWindow::playSingleVideo(std::string ytid)
    {
        // Запуск mpv в заднем фоне
        std::string yt_url = "https://youtu.be/"+ytid;

        Glib::spawn_async(
            "",
            m_appcore->getMPVLaunchParams(ytid),
            Glib::SpawnFlags::DEFAULT
        );
    }

    void MainWindow::searchSidebarRegularMode()
    {
        // Получить иконку поиска и создать надпись "Поиск"
        auto box = Gtk::make_managed<Gtk::Box>(Gtk::Orientation::VERTICAL);
        box->set_valign(Gtk::Align::CENTER);

        auto icon = Gio::Icon::create("find");
        auto img_icon = Gtk::make_managed<Gtk::Image>(icon);
        auto label = Gtk::make_managed<Gtk::Label>("Search results");

        box->append(*img_icon);
        box->append(*label);

        m_search_body->set_start_child(*box);
    }

    void MainWindow::searchSidebarError(Glib::ustring description)
    {
        auto builder = Gtk::Builder::create_from_resource(
            "/src/YTMPV/res/sidebar/error.ui");
        Gtk::Box* layout = builder->get_widget<Gtk::Box>("layout");
        Gtk::Label* label = builder->get_widget<Gtk::Label>("errorLabel");

        assert(layout != nullptr);
        assert(label != nullptr);

        label->set_label(Glib::Markup::escape_text(description));
        m_search_body->set_start_child(*layout);
    }

    void MainWindow::searchSidebarVideoDetailsMode(
        Glib::ustring title,
        Glib::ustring channel_title,
        Glib::ustring description,
        std::string published_at,
        std::string yt_id,
        int views_count,
        std::string thumbnail_url)
    {
        // Главный объект
        auto layout = Gtk::make_managed<Gtk::Box>(Gtk::Orientation::VERTICAL);
        // Прокручиваемое окно
        auto scroller = Gtk::make_managed<Gtk::ScrolledWindow>();
        // Область деталей
        auto box_details = Gtk::make_managed<Gtk::Box>(Gtk::Orientation::VERTICAL);
        // Область деталей (текст)
        auto box_textdetails = Gtk::make_managed<Gtk::Box>(Gtk::Orientation::VERTICAL);
        // Область кнопок
        auto box_buttons = Gtk::make_managed<Gtk::Box>(Gtk::Orientation::VERTICAL);
        
        auto thumbnail = Gtk::make_managed<Gtk::Picture>();
        auto lbl_title = Gtk::make_managed<Gtk::Label>();
        auto lbl_channel_title = Gtk::make_managed<Gtk::Label>(channel_title);
        auto lbl_views_count = Gtk::make_managed<Gtk::Label>();
        auto lbl_description = Gtk::make_managed<Gtk::Label>(description);
        auto lbl_published_at = Gtk::make_managed<Gtk::Label>(published_at);

        // Установка названия видео (выделение полужирным)
        Glib::ustring title_escaped = Glib::Markup::escape_text(title);
        lbl_title->set_markup("<b>"+title_escaped+"</b>");

        // Количество просмотров
        lbl_views_count->set_text(std::to_string(views_count) + " " + "views");

        // Добавление кнопки "Воспроизвести"
        auto btn_play = Gtk::make_managed<Gtk::Button>("Play video");
        btn_play->set_hexpand();
        btn_play->signal_clicked().connect(sigc::bind(
            sigc::mem_fun(*this, &MainWindow::onPlayButtonClicked),
            PlayButtonParams {yt_id, btn_play}
        ));
        box_buttons->append(*btn_play);

        // Кнопки оценки видео
        int my_rating = db::getVideoRating(yt_id);
        updateVideoDetails(yt_id, static_cast<core::VideoRating>(my_rating));

        // Добавление кнопки "Открыть в веб-браузере"
        auto btn_webbrowser = Gtk::make_managed<Gtk::Button>("Open in web browser");
        btn_webbrowser->set_hexpand();
        btn_webbrowser->signal_clicked().connect(sigc::bind(
            sigc::mem_fun(*this, &MainWindow::openVideoInWebBrowser),
            yt_id
        ));

        // Скачивание большого превью
        std::string thumbnail_path = "thumbnails/large/" + yt_id + ".jpg";
        Glib::RefPtr<Gdk::Pixbuf> thumbnail_pixbuf;
        try {
            thumbnail_pixbuf = Gdk::Pixbuf::create_from_file(thumbnail_path);
        } catch (Glib::FileError& e) {
            switch (e.code()) {
                case Glib::FileError::Code::NO_SUCH_ENTITY:
                    // Превью на диске не найдено. Скачиваем...
                    ytapi::downloadThumbnail(thumbnail_url, thumbnail_path);
                    // ...и присваиваем
                    thumbnail_pixbuf = Gdk::Pixbuf::create_from_file(thumbnail_path);
                    break;

                default:
                    std::cout << "Unhandled error when creating thumbnail. Code is: " << e.code(); 
                    break;
            }
        }
        auto thumbnail_texture = Gdk::Texture::create_for_pixbuf(thumbnail_pixbuf);
        thumbnail->set_paintable(thumbnail_texture);
        thumbnail->set_content_fit(Gtk::ContentFit::COVER);
        thumbnail->set_hexpand();
        thumbnail->set_size_request(-1, 240);

        lbl_description->set_halign(Gtk::Align::START);
        lbl_description->set_selectable();
        lbl_description->set_wrap();

        box_textdetails->append(*lbl_title);
        box_textdetails->append(*lbl_channel_title);
        box_textdetails->append(*lbl_views_count);
        box_textdetails->append(*lbl_published_at);
        m_rate_container.unparent();
        box_textdetails->append(m_rate_container);
        box_textdetails->append(*btn_webbrowser);
        box_textdetails->append(*lbl_description);
        box_textdetails->set_margin(8);

        box_details->append(*thumbnail);
        box_details->append(*box_textdetails);
        box_details->set_vexpand();
        scroller->set_child(*box_details);

        layout->append(*scroller);
        layout->append(*box_buttons);

        m_search_body->set_start_child(*layout);
    }

    void MainWindow::openVideoInWebBrowser(std::string yt_id)
    {
        // TODO:
        // - Windows
        // - MacOs
        std::string command = "xdg-open https://youtube.com/watch?v="+yt_id;
        system(command.c_str());
    }

    void MainWindow::onPlayButtonClicked(PlayButtonParams e)
    {
        e.caller->set_sensitive(false);
        e.caller->set_label("MPV started");
        MainWindow::playSingleVideo(e.yt_id);
    }

    void MainWindow::onRateButtonClicked(RateButtonParams e)
    {
        std::string current_ytid = m_appcore->getSelectedYTID();
        updateVideoDetails(current_ytid, e.rating);
        m_appcore->rateVideo(current_ytid, e.rating);
    }

    void MainWindow::updateVideoDetails(std::string ytid, core::VideoRating rating)
    {
        m_appcore->setSelectedYTID(ytid);
        m_appcore->setCurrentRating(rating);
        
        m_rate_dislike_btn->set_state_flags(Gtk::StateFlags::NORMAL, true);
        m_rate_like_btn->set_state_flags(Gtk::StateFlags::NORMAL, true);
        m_rate_meh_btn->set_state_flags(Gtk::StateFlags::NORMAL, true);
        switch (rating)
        {
            case core::VideoRating::NONE:
                m_rate_meh_btn->set_state_flags(Gtk::StateFlags::CHECKED);
                break;
            case core::VideoRating::LIKE:
                m_rate_like_btn->set_state_flags(Gtk::StateFlags::CHECKED);
                break;
            case core::VideoRating::DISLIKE:
                m_rate_dislike_btn->set_state_flags(Gtk::StateFlags::CHECKED);
                break;
            default:
                break;
        }
    }
}
