#include "PrefWindow.hpp"
#include "core.hpp"
#include <array>
#include <fstream>
#include <iostream>
#include <gtkmm/stringlist.h>
#include <assert.h>
#include <nlohmann/json.hpp>
#include "app.hpp"
#include <YoutubeApi/youtubeapi.hpp>

using json = nlohmann::json;

namespace components {
    PrefWindow::PrefWindow(
        BaseObjectType* cobject,
        const Glib::RefPtr<Gtk::Builder>& refBuilder,
        core::Core* appcore)
        // init
        : Gtk::Window(cobject)
        , m_appcore(appcore)
        , m_builder(refBuilder)
    {
        // Заполнение выпадающего списка качества видео
        std::vector<Glib::ustring> quality_strings =
        {
            "240p",
            "360p",
            "480p",
            "720p",
            "1080p"
        };

        m_save_btn          = m_builder->get_widget<Gtk::Button>("saveclose_button");
        m_cancel_btn        = m_builder->get_widget<Gtk::Button>("cancel_button");
        m_mpvpath_entry     = m_builder->get_widget<Gtk::Entry>("mpvpath_entry");
        m_quality_dropdown  = m_builder->get_widget<Gtk::DropDown>("quality_entry");
        m_apikey_entry      = m_builder->get_widget<Gtk::Entry>("apikey_entry");

        m_miniplayer_btn    = m_builder->get_widget<Gtk::CheckButton>("minimode_entry");
        m_fullscreen_btn    = m_builder->get_widget<Gtk::CheckButton>("fullscreen_entry");
        m_ontop_btn         = m_builder->get_widget<Gtk::CheckButton>("ontop_entry");

        assert(m_quality_dropdown   != nullptr);
        assert(m_cancel_btn         != nullptr);
        assert(m_save_btn           != nullptr);
        assert(m_mpvpath_entry      != nullptr);
        assert(m_apikey_entry       != nullptr);
        assert(m_miniplayer_btn     != nullptr);
        assert(m_fullscreen_btn     != nullptr);
        assert(m_ontop_btn          != nullptr);

        // Кнопки
        m_save_btn->signal_clicked().connect(sigc::mem_fun(
            *this, &PrefWindow::onSaveClicked));
        m_cancel_btn->signal_clicked().connect(sigc::mem_fun(
            *this, &PrefWindow::onCancelClicked));

        // Ключ API
        m_apikey_entry->set_text(appcore->getApiKey());

        // Путь к MPV
        m_mpvpath_entry->set_text(appcore->getMpvPath());

        // Качество
        auto string_list = Gtk::StringList::create(quality_strings);
        m_quality_dropdown->set_model(string_list);
        int selected_quality_index;
        switch(m_appcore->getQuality())
        {
            case 240:
                selected_quality_index = 0;
                break;
            case 360:
                selected_quality_index = 1;
                break;
            case 480:
                selected_quality_index = 2;
                break;
            case 720:
                selected_quality_index = 3;
                break;
            case 1080:
                selected_quality_index = 4;
                break;
            default:
                selected_quality_index = 0;
                break;
        }
        m_quality_dropdown->set_selected(selected_quality_index);

        // Мини-плеер
        this->prepareBoolButton(
            m_miniplayer_btn,
            appcore->getIsMiniplayer(),
            BoolSettingType::MINIPLAYER);

        // Полноэкранный режим
        this->prepareBoolButton(
            m_fullscreen_btn,
            appcore->getFullscreen(),
            BoolSettingType::FULLSCREEN);

        // Режим поверх всех окон
        this->prepareBoolButton(
            m_ontop_btn,
            appcore->getOnTop(),
            BoolSettingType::ONTOP);
    }

    PrefWindow* PrefWindow::create(
        Gtk::Window& parent,
        core::Core* appcore)
    {
        auto builder = Gtk::Builder::create_from_resource("/src/YTMPV/res/PrefsWindow.ui");
        auto win = Gtk::Builder::get_widget_derived<PrefWindow>(
            builder,
            "prefs_dialog",
            appcore);
        assert(win != nullptr);

        win->set_transient_for(parent);
        return win;
    }

    void PrefWindow::onCancelClicked()
    {
        this->close();
    }

    void PrefWindow::onSaveClicked()
    {
        this->saveConfig();
        this->close();
    }

    void PrefWindow::onBoolButtonClicked(BoolButtonParams params)
    {
        bool value = params.caller->get_active();
        switch (params.setting_type)
        {
            case BoolSettingType::MINIPLAYER:
                m_appcore->setMiniplayer(value);
                break;
            case BoolSettingType::FULLSCREEN:
                m_appcore->setFullscreen(value);
                break;
            case BoolSettingType::ONTOP:
                m_appcore->setOnTop(value);
                break;
            default:
                break;
        }
    }

    void PrefWindow::saveConfig()
    {
        // Качество
        int quality;
        int sel_index = m_quality_dropdown->get_selected();
        switch(sel_index)
        {
            case 0:
                quality = 240;
                break;
            case 1:
                quality = 360;
                break;
            case 2:
                quality = 480;
                break;
            case 3:
                quality = 720;
                break;
            case 4:
                quality = 1080;
                break;
            default:
                quality = 240;
                break;
        }
        m_appcore->setQuality(quality);
        m_appcore->setApiKey(m_apikey_entry->get_text());

        m_appcore->saveConfig();
    }

    void PrefWindow::prepareBoolButton(
            Gtk::CheckButton* widget,
            bool default_value,
            BoolSettingType setting_type)
    {
        widget->set_active(default_value);
        widget->signal_toggled().connect(
            sigc::bind(
                sigc::mem_fun(*this, &PrefWindow::onBoolButtonClicked),
                BoolButtonParams {
                    setting_type,
                    widget
                }
            )
        );
    }
}
