#include "core.hpp"
#include <nlohmann/json.hpp>
#include <Database/database.hpp>
#include <iomanip>

using json = nlohmann::json;

namespace core {

    Core::Core(
        std::string apikey,
        std::string config_path,
        int quality,
        bool launch_fullscreen,
        bool launch_ontop,
        bool launch_miniplayer,
        std::string mpvpath,
        std::string thumbnails_dir)
        // Инициализация
        : m_apikey(apikey)
        , m_selected_ytid("")
        , m_config_path(config_path)
        , m_quality(quality)
        , m_launch_fullscreen(launch_fullscreen)
        , m_launch_ontop(launch_ontop)
        , m_launch_miniplayer(launch_miniplayer)
        , m_mpvpath(mpvpath)
        , m_thumbnails_dir(thumbnails_dir)
    {}

    void Core::setQuality(int quality)
    {
        this->m_quality = quality;
    }

    void Core::saveConfig()
    {
        // Сериализация
        json j_object = {
            {"quality",     this->m_quality},
            {"fullscreen",  this->m_launch_fullscreen},
            {"ontop",       this->m_launch_ontop},
            {"miniplayer",  this->m_launch_miniplayer},
            {"apikey",      this->m_apikey},
            {"mpvpath",     this->m_mpvpath},
        };
        std::string dump = j_object.dump(1, '\t');

        // Запись в файл
        std::ofstream config(m_config_path);
        config << dump;
        config.close();
    }

    db::video Core::getVideoByYTID(std::string ytid)
    {
        db::video output;

        // 1. Проверить есть ли видео в базе данных
        if (db::getVideoByYTID(ytid, &output)) {
            return output;
        }

        // Видео в базе данных нет. Получаем информацию о нём через ytapi
        ytapi::Video found_video = ytapi::getVideoByYTID(ytid, m_apikey);
        output.ytid         = ytid;
        output.title        = found_video.title;
        output.description  = found_video.description;
        output.published_at = found_video.published_at;
        output.views_count  = found_video.views_count;
        output.rate_status  = found_video.rating;
        output.big_thumbnail= found_video.big_thumbnail;

        // Ищем автора в базе данных
        db::author video_author;
        if (!db::getAuthorByYTID(found_video.author_yt_id, &video_author)) {
            // Автора в базе данных нет. Добавляем
            video_author.ytid = found_video.author_yt_id;
            video_author.name = found_video.author_name;
            db::addAuthor(&video_author);
        }
        output.author_obj = video_author;

        // Добавляем видео в БД
        db::addVideo(&output);

        return output;
    }

    std::string Core::getYTDLFormat()
    {
        return
            "--ytdl-format=bestvideo[height<=?"
            +std::to_string(m_quality)
            +"]+bestaudio/best";
    }

    // Устанавливает оценку в БД и посылает запрос на оценку
    void Core::rateVideo(std::string ytid, VideoRating rating)
    {
        // TODO: ничего не делать если текущая оценка такая же как и rating
        db::setVideoRating(ytid, static_cast<int>(rating));
        /*ytapi::setVideoRating(ytid, static_cast<int>(rating), m_client_secret);*/
    }

    std::vector<std::string> Core::getMPVLaunchParams(std::string ytid)
    {
        std::vector<std::string> output;

        // Путь до MPV
        output.push_back(m_mpvpath);

        // Ссылка на видео
        output.push_back(getYTLink(ytid));

        // Без терминала
        output.push_back("--no-terminal");

        // С качеством:
        output.push_back(getYTDLFormat());

        // -- Параметры окна --
        if (m_launch_miniplayer) {
            output.push_back("--geometry=640x480-0-0");
            output.push_back("--ontop");
        } else {
            if (m_launch_fullscreen) {
                output.push_back("--fullscreen");
            }
            if (m_launch_ontop) {
                output.push_back("--ontop");
            }
        }

        return output;
    }

    std::string Core::getSelectedYTID()
    {
        return m_selected_ytid;
    }

    std::string Core::getSelectedYTLink()
    {
        return getYTLink(m_selected_ytid);
    }

    std::string Core::getYTLink(std::string ytid)
    {
        return "https://youtube.com/watch?v=" + ytid;
    }

    void Core::setSelectedYTID(std::string ytid)
    {
        m_selected_ytid = ytid;
    }

    void Core::setCurrentRating(VideoRating rating)
    {
        m_current_rating = rating;
    }

    int Core::getQuality()
    {
        return m_quality;
    }

    ytapi::SearchResults Core::searchVideos(std::string query)
    {
        return ytapi::getVideoIDsByQuery(query, m_apikey);
    }

    void Core::setMpvPath(std::string path)
    {
        m_mpvpath = path;
    }

    std::string Core::getApiKey()
    {
        return m_apikey;
    }

    std::string Core::getMpvPath()
    {
        return m_mpvpath;
    }

    bool Core::getIsMiniplayer()
    {
        return m_launch_miniplayer;
    }

    void Core::setMiniplayer(bool value)
    {
        m_launch_miniplayer = value;
    }

    void Core::setFullscreen(bool value)
    {
        m_launch_fullscreen = value;
    }

    void Core::setOnTop(bool value)
    {
        m_launch_ontop = value;
    }

    bool Core::getFullscreen()
    {
        return m_launch_fullscreen;
    }

    bool Core::getOnTop()
    {
        return m_launch_ontop;
    }

    void Core::setApiKey(std::string value)
    {
        m_apikey = value;
    }

    std::string Core::getNumberGrouped(int n)
    {
        float value;
        std::string suffix;
        
        if (n > 1000000) {
            value = (float)n / 1000000.0;
            suffix = "M";

        } else if (n > 1000) {
            value = (float)n / 1000.0;
            suffix = "K";

        } else {
            value = n;
            suffix = "";
        } 

        std::stringstream value_stream;
        value_stream << std::fixed << std::setprecision(3) << value;

        return value_stream.str() + suffix;
    }

    std::string Core::getDateText(std::string date)
    {
        // 1. Получить информацию о времени
        std::tm t;
        std::istringstream ss(date);
        ss >> std::get_time(&t, "%Y-%m-%dT%H:%M:%SZ");
        if (ss.fail()) {
            return "<couldn't format date>";
        }

        // 2. Сконструировать текст
        std::stringstream output;
        output << std::put_time(&t, "%c");
        return output.str();
    }
}
